@@ -19,6 +19,7 @@ from . import mlog
 from .compilers import *
 from .mesonlib import EnvironmentException, Popen_safe
 import configparser
+import shlex
 import shutil
 
 build_filename = 'meson.build'
@@ -121,6 +122,18 @@ def detect_cpu_family(compilers):
     if trial.startswith('arm'):
         return 'arm'
     if trial in ('amd64', 'x64'):
+        trial = 'x86_64'
+    if trial == 'x86_64':
+        # On Linux (and maybe others) there can be any mixture of 32/64 bit
+        # code in the kernel, Python, system etc. The only reliable way
+        # to know is to check the compiler defines.
+        for c in compilers.values():
+            try:
+                if c.has_define('__i386__'):
+                    return 'x86'
+            except mesonlib.MesonException:
+                # Ignore compilers that do not support has_define.
+                pass
         return 'x86_64'
     # Add fixes here as bugs are reported.
     return trial
@@ -131,6 +144,15 @@ def detect_cpu(compilers):
     else:
         trial = platform.machine().lower()
     if trial in ('amd64', 'x64'):
+        trial = 'x86_64'
+    if trial == 'x86_64':
+        # Same check as above for cpu_family
+        for c in compilers.values():
+            try:
+                if c.has_define('__i386__'):
+                    return 'i686' # All 64 bit cpus have at least this level of x86 support.
+            except mesonlib.MesonException:
+                pass
         return 'x86_64'
     # Add fixes here as bugs are reported.
     return trial
@@ -239,6 +261,10 @@ class Environment:
             self.exe_suffix = ''
             self.object_suffix = 'o'
             self.win_libdir_layout = False
+        if 'STRIP' in os.environ:
+            self.native_strip_bin = shlex.split('STRIP')
+        else:
+            self.native_strip_bin = ['strip']
 
     def is_cross_build(self):
         return self.cross_info is not None
@@ -345,10 +371,13 @@ class Environment:
         # We ignore Cygwin for now, and treat it as a standard GCC
         return GCC_STANDARD
 
-    def detect_c_compiler(self, want_cross):
-        evar = 'CC'
+    def _get_compilers(self, lang, evar, want_cross):
+        '''
+        The list of compilers is detected in the exact same way for
+        C, C++, ObjC, ObjC++, Fortran so consolidate it here.
+        '''
         if self.is_cross_build() and want_cross:
-            compilers = [self.cross_info.config['binaries']['c']]
+            compilers = [mesonlib.stringlistify(self.cross_info.config['binaries'][lang])]
             ccache = []
             is_cross = True
             if self.cross_info.need_exe_wrapper():
@@ -356,122 +385,122 @@ class Environment:
             else:
                 exe_wrap = []
         elif evar in os.environ:
-            compilers = os.environ[evar].split()
+            compilers = [shlex.split(os.environ[evar])]
             ccache = []
             is_cross = False
             exe_wrap = None
         else:
-            compilers = self.default_c
+            compilers = getattr(self, 'default_' + lang)
             ccache = self.detect_ccache()
             is_cross = False
             exe_wrap = None
+        return compilers, ccache, is_cross, exe_wrap
+
+    def _handle_compiler_exceptions(self, exceptions, compilers):
+        errmsg = 'Unknown compiler(s): ' + str(compilers)
+        if exceptions:
+            errmsg += '\nThe follow exceptions were encountered:'
+            for (c, e) in exceptions.items():
+                errmsg += '\nRunning "{0}" gave "{1}"'.format(c, e)
+        raise EnvironmentException(errmsg)
+
+    def _detect_c_or_cpp_compiler(self, lang, evar, want_cross):
         popen_exceptions = {}
+        compilers, ccache, is_cross, exe_wrap = self._get_compilers(lang, evar, want_cross)
         for compiler in compilers:
+            if isinstance(compiler, str):
+                compiler = [compiler]
             try:
-                basename = os.path.basename(compiler).lower()
-                if basename == 'cl' or basename == 'cl.exe':
+                if 'cl' in compiler or 'cl.exe' in compiler:
                     arg = '/?'
                 else:
                     arg = '--version'
-                p, out, err = Popen_safe([compiler, arg])
+                p, out, err = Popen_safe(compiler + [arg])
             except OSError as e:
-                popen_exceptions[' '.join([compiler, arg])] = e
+                popen_exceptions[' '.join(compiler + [arg])] = e
                 continue
             version = search_version(out)
             if 'Free Software Foundation' in out:
-                defines = self.get_gnu_compiler_defines([compiler])
+                defines = self.get_gnu_compiler_defines(compiler)
                 if not defines:
                     popen_exceptions[compiler] = 'no pre-processor defines'
                     continue
                 gtype = self.get_gnu_compiler_type(defines)
                 version = self.get_gnu_version_from_defines(defines)
-                return GnuCCompiler(ccache + [compiler], version, gtype, is_cross, exe_wrap, defines)
+                cls = GnuCCompiler if lang == 'c' else GnuCPPCompiler
+                return cls(ccache + compiler, version, gtype, is_cross, exe_wrap, defines)
             if 'clang' in out:
                 if 'Apple' in out or for_darwin(want_cross, self):
                     cltype = CLANG_OSX
                 else:
                     cltype = CLANG_STANDARD
-                return ClangCCompiler(ccache + [compiler], version, cltype, is_cross, exe_wrap)
+                cls = ClangCCompiler if lang == 'c' else ClangCPPCompiler
+                return cls(ccache + compiler, version, cltype, is_cross, exe_wrap)
             if 'Microsoft' in out or 'Microsoft' in err:
                 # Visual Studio prints version number to stderr but
                 # everything else to stdout. Why? Lord only knows.
                 version = search_version(err)
-                return VisualStudioCCompiler([compiler], version, is_cross, exe_wrap)
+                cls = VisualStudioCCompiler if lang == 'c' else VisualStudioCPPCompiler
+                return cls(compiler, version, is_cross, exe_wrap)
             if '(ICC)' in out:
                 # TODO: add microsoft add check OSX
                 inteltype = ICC_STANDARD
-                return IntelCCompiler(ccache + [compiler], version, inteltype, is_cross, exe_wrap)
-        errmsg = 'Unknown compiler(s): "' + ', '.join(compilers) + '"'
-        if popen_exceptions:
-            errmsg += '\nThe follow exceptions were encountered:'
-            for (c, e) in popen_exceptions.items():
-                errmsg += '\nRunning "{0}" gave "{1}"'.format(c, e)
-        raise EnvironmentException(errmsg)
+                cls = IntelCCompiler if lang == 'c' else IntelCPPCompiler
+                return cls(ccache + compiler, version, inteltype, is_cross, exe_wrap)
+        self._handle_compiler_exceptions(popen_exceptions, compilers)
+
+    def detect_c_compiler(self, want_cross):
+        return self._detect_c_or_cpp_compiler('c', 'CC', want_cross)
+
+    def detect_cpp_compiler(self, want_cross):
+        return self._detect_c_or_cpp_compiler('cpp', 'CXX', want_cross)
 
     def detect_fortran_compiler(self, want_cross):
-        evar = 'FC'
-        if self.is_cross_build() and want_cross:
-            compilers = [self.cross_info['fortran']]
-            is_cross = True
-            if self.cross_info.need_exe_wrapper():
-                exe_wrap = self.cross_info.get('exe_wrapper', None)
-            else:
-                exe_wrap = []
-        elif evar in os.environ:
-            compilers = os.environ[evar].split()
-            is_cross = False
-            exe_wrap = None
-        else:
-            compilers = self.default_fortran
-            is_cross = False
-            exe_wrap = None
         popen_exceptions = {}
+        compilers, ccache, is_cross, exe_wrap = self._get_compilers('fortran', 'FC', want_cross)
         for compiler in compilers:
+            if isinstance(compiler, str):
+                compiler = [compiler]
             for arg in ['--version', '-V']:
                 try:
-                    p, out, err = Popen_safe([compiler, arg])
+                    p, out, err = Popen_safe(compiler + [arg])
                 except OSError as e:
-                    popen_exceptions[' '.join([compiler, arg])] = e
+                    popen_exceptions[' '.join(compiler + [arg])] = e
                     continue
 
                 version = search_version(out)
 
                 if 'GNU Fortran' in out:
-                    defines = self.get_gnu_compiler_defines([compiler])
+                    defines = self.get_gnu_compiler_defines(compiler)
                     if not defines:
                         popen_exceptions[compiler] = 'no pre-processor defines'
                         continue
                     gtype = self.get_gnu_compiler_type(defines)
                     version = self.get_gnu_version_from_defines(defines)
-                    return GnuFortranCompiler([compiler], version, gtype, is_cross, exe_wrap, defines)
+                    return GnuFortranCompiler(compiler, version, gtype, is_cross, exe_wrap, defines)
 
                 if 'G95' in out:
-                    return G95FortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return G95FortranCompiler(compiler, version, is_cross, exe_wrap)
 
                 if 'Sun Fortran' in err:
                     version = search_version(err)
-                    return SunFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return SunFortranCompiler(compiler, version, is_cross, exe_wrap)
 
                 if 'ifort (IFORT)' in out:
-                    return IntelFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return IntelFortranCompiler(compiler, version, is_cross, exe_wrap)
 
                 if 'PathScale EKOPath(tm)' in err:
-                    return PathScaleFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return PathScaleFortranCompiler(compiler, version, is_cross, exe_wrap)
 
                 if 'PGI Compilers' in out:
-                    return PGIFortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return PGIFortranCompiler(compiler, version, is_cross, exe_wrap)
 
                 if 'Open64 Compiler Suite' in err:
-                    return Open64FortranCompiler([compiler], version, is_cross, exe_wrap)
+                    return Open64FortranCompiler(compiler, version, is_cross, exe_wrap)
 
                 if 'NAG Fortran' in err:
-                    return NAGFortranCompiler([compiler], version, is_cross, exe_wrap)
-        errmsg = 'Unknown compiler(s): "' + ', '.join(compilers) + '"'
-        if popen_exceptions:
-            errmsg += '\nThe follow exceptions were encountered:'
-            for (c, e) in popen_exceptions.items():
-                errmsg += '\nRunning "{0}" gave "{1}"'.format(c, e)
-        raise EnvironmentException(errmsg)
+                    return NAGFortranCompiler(compiler, version, is_cross, exe_wrap)
+        self._handle_compiler_exceptions(popen_exceptions, compilers)
 
     def get_scratch_dir(self):
         return self.scratch_dir
@@ -480,115 +509,57 @@ class Environment:
         path = os.path.split(__file__)[0]
         return os.path.join(path, 'depfixer.py')
 
-    def detect_cpp_compiler(self, want_cross):
-        evar = 'CXX'
-        if self.is_cross_build() and want_cross:
-            compilers = [self.cross_info.config['binaries']['cpp']]
-            ccache = []
-            is_cross = True
-            if self.cross_info.need_exe_wrapper():
-                exe_wrap = self.cross_info.config['binaries'].get('exe_wrapper', None)
-            else:
-                exe_wrap = []
-        elif evar in os.environ:
-            compilers = os.environ[evar].split()
-            ccache = []
-            is_cross = False
-            exe_wrap = None
-        else:
-            compilers = self.default_cpp
-            ccache = self.detect_ccache()
-            is_cross = False
-            exe_wrap = None
+    def detect_objc_compiler(self, want_cross):
         popen_exceptions = {}
+        compilers, ccache, is_cross, exe_wrap = self._get_compilers('objc', 'OBJC', want_cross)
         for compiler in compilers:
-            basename = os.path.basename(compiler).lower()
-            if basename == 'cl' or basename == 'cl.exe':
-                arg = '/?'
-            else:
-                arg = '--version'
+            if isinstance(compiler, str):
+                compiler = [compiler]
+            arg = ['--version']
             try:
-                p, out, err = Popen_safe([compiler, arg])
+                p, out, err = Popen_safe(compiler + arg)
             except OSError as e:
-                popen_exceptions[' '.join([compiler, arg])] = e
-                continue
+                popen_exceptions[' '.join(compiler + arg)] = e
             version = search_version(out)
             if 'Free Software Foundation' in out:
-                defines = self.get_gnu_compiler_defines([compiler])
+                defines = self.get_gnu_compiler_defines(compiler)
                 if not defines:
                     popen_exceptions[compiler] = 'no pre-processor defines'
                     continue
                 gtype = self.get_gnu_compiler_type(defines)
                 version = self.get_gnu_version_from_defines(defines)
-                return GnuCPPCompiler(ccache + [compiler], version, gtype, is_cross, exe_wrap, defines)
-            if 'clang' in out:
-                if 'Apple' in out:
-                    cltype = CLANG_OSX
-                else:
-                    cltype = CLANG_STANDARD
-                return ClangCPPCompiler(ccache + [compiler], version, cltype, is_cross, exe_wrap)
-            if 'Microsoft' in out or 'Microsoft' in err:
-                version = search_version(err)
-                return VisualStudioCPPCompiler([compiler], version, is_cross, exe_wrap)
-            if '(ICC)' in out:
-                # TODO: add microsoft add check OSX
-                inteltype = ICC_STANDARD
-                return IntelCPPCompiler(ccache + [compiler], version, inteltype, is_cross, exe_wrap)
-        errmsg = 'Unknown compiler(s): "' + ', '.join(compilers) + '"'
-        if popen_exceptions:
-            errmsg += '\nThe follow exceptions were encountered:'
-            for (c, e) in popen_exceptions.items():
-                errmsg += '\nRunning "{0}" gave "{1}"'.format(c, e)
-        raise EnvironmentException(errmsg)
-
-    def detect_objc_compiler(self, want_cross):
-        if self.is_cross_build() and want_cross:
-            exelist = [self.cross_info['objc']]
-            is_cross = True
-            if self.cross_info.need_exe_wrapper():
-                exe_wrap = self.cross_info.get('exe_wrapper', None)
-            else:
-                exe_wrap = []
-        else:
-            exelist = self.get_objc_compiler_exelist()
-            is_cross = False
-            exe_wrap = None
-        try:
-            p, out, err = Popen_safe(exelist + ['--version'])
-        except OSError:
-            raise EnvironmentException('Could not execute ObjC compiler "%s"' % ' '.join(exelist))
-        version = search_version(out)
-        if 'Free Software Foundation' in out:
-            defines = self.get_gnu_compiler_defines(exelist)
-            version = self.get_gnu_version_from_defines(defines)
-            return GnuObjCCompiler(exelist, version, is_cross, exe_wrap, defines)
-        if out.startswith('Apple LLVM'):
-            return ClangObjCCompiler(exelist, version, CLANG_OSX, is_cross, exe_wrap)
-        raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+                return GnuObjCCompiler(ccache + compiler, version, gtype, is_cross, exe_wrap, defines)
+            if out.startswith('Apple LLVM'):
+                return ClangObjCCompiler(ccache + compiler, version, CLANG_OSX, is_cross, exe_wrap)
+            if out.startswith('clang'):
+                return ClangObjCCompiler(ccache + compiler, version, CLANG_STANDARD, is_cross, exe_wrap)
+        self._handle_compiler_exceptions(popen_exceptions, compilers)
 
     def detect_objcpp_compiler(self, want_cross):
-        if self.is_cross_build() and want_cross:
-            exelist = [self.cross_info['objcpp']]
-            is_cross = True
-            if self.cross_info.need_exe_wrapper():
-                exe_wrap = self.cross_info.get('exe_wrapper', None)
-            else:
-                exe_wrap = []
-        else:
-            exelist = self.get_objcpp_compiler_exelist()
-            is_cross = False
-            exe_wrap = None
-        try:
-            p, out, err = Popen_safe(exelist + ['--version'])
-        except OSError:
-            raise EnvironmentException('Could not execute ObjC++ compiler "%s"' % ' '.join(exelist))
-        version = search_version(out)
-        if 'Free Software Foundation' in out:
-            defines = self.get_gnu_compiler_defines(exelist)
-            return GnuObjCPPCompiler(exelist, version, is_cross, exe_wrap, defines)
-        if out.startswith('Apple LLVM'):
-            return ClangObjCPPCompiler(exelist, version, CLANG_OSX, is_cross, exe_wrap)
-        raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
+        popen_exceptions = {}
+        compilers, ccache, is_cross, exe_wrap = self._get_compilers('objcpp', 'OBJCXX', want_cross)
+        for compiler in compilers:
+            if isinstance(compiler, str):
+                compiler = [compiler]
+            arg = ['--version']
+            try:
+                p, out, err = Popen_safe(compiler + arg)
+            except OSError as e:
+                popen_exceptions[' '.join(compiler + arg)] = e
+            version = search_version(out)
+            if 'Free Software Foundation' in out:
+                defines = self.get_gnu_compiler_defines(compiler)
+                if not defines:
+                    popen_exceptions[compiler] = 'no pre-processor defines'
+                    continue
+                gtype = self.get_gnu_compiler_type(defines)
+                version = self.get_gnu_version_from_defines(defines)
+                return GnuObjCPPCompiler(ccache + compiler, version, gtype, is_cross, exe_wrap, defines)
+            if out.startswith('Apple LLVM'):
+                return ClangObjCPPCompiler(ccache + compiler, version, CLANG_OSX, is_cross, exe_wrap)
+            if out.startswith('clang'):
+                return ClangObjCPPCompiler(ccache + compiler, version, CLANG_STANDARD, is_cross, exe_wrap)
+        self._handle_compiler_exceptions(popen_exceptions, compilers)
 
     def detect_java_compiler(self):
         exelist = ['javac']
@@ -641,9 +612,9 @@ class Environment:
         # environment variable because LDC has a much more
         # up to date language version at time (2016).
         if 'DC' in os.environ:
-            exelist = os.environ['DC'].split()
+            exelist = shlex.split(os.environ['DC'])
         elif self.is_cross_build() and want_cross:
-            exelist = [self.cross_info.config['binaries']['d']]
+            exelist = mesonlib.stringlistify(self.cross_info.config['binaries']['d'])
             is_cross = True
         elif shutil.which("ldc2"):
             exelist = ['ldc2']
@@ -683,30 +654,31 @@ class Environment:
     def detect_static_linker(self, compiler):
         if compiler.is_cross:
             linker = self.cross_info.config['binaries']['ar']
+            if isinstance(linker, str):
+                linker = [linker]
         else:
             evar = 'AR'
             if evar in os.environ:
-                linker = os.environ[evar].strip()
+                linker = shlex.split(os.environ[evar])
             elif isinstance(compiler, VisualStudioCCompiler):
-                linker = self.vs_static_linker
+                linker = [self.vs_static_linker]
             else:
-                linker = self.default_static_linker
-        basename = os.path.basename(linker).lower()
-        if basename == 'lib' or basename == 'lib.exe':
+                linker = [self.default_static_linker]
+        if 'lib' in linker or 'lib.exe' in linker:
             arg = '/?'
         else:
             arg = '--version'
         try:
-            p, out, err = Popen_safe([linker, arg])
+            p, out, err = Popen_safe(linker + [arg])
         except OSError:
-            raise EnvironmentException('Could not execute static linker "%s".' % linker)
+            raise EnvironmentException('Could not execute static linker "%s".' % ' '.join(linker))
         if '/OUT:' in out or '/OUT:' in err:
-            return VisualStudioLinker([linker])
+            return VisualStudioLinker(linker)
         if p.returncode == 0:
-            return ArLinker([linker])
+            return ArLinker(linker)
         if p.returncode == 1 and err.startswith('usage'): # OSX
-            return ArLinker([linker])
-        raise EnvironmentException('Unknown static linker "%s"' % linker)
+            return ArLinker(linker)
+        raise EnvironmentException('Unknown static linker "%s"' % ' '.join(linker))
 
     def detect_ccache(self):
         try:
@@ -719,20 +691,6 @@ class Environment:
             cmdlist = []
         return cmdlist
 
-    def get_objc_compiler_exelist(self):
-        ccachelist = self.detect_ccache()
-        evar = 'OBJCC'
-        if evar in os.environ:
-            return os.environ[evar].split()
-        return ccachelist + self.default_objc
-
-    def get_objcpp_compiler_exelist(self):
-        ccachelist = self.detect_ccache()
-        evar = 'OBJCXX'
-        if evar in os.environ:
-            return os.environ[evar].split()
-        return ccachelist + self.default_objcpp
-
     def get_source_dir(self):
         return self.source_dir
 
